# -*- coding: utf-8 -*-
"""
Class for storing and handling partitions in layers.

Authors:
Alice Patania <alice.patania@uvm.edu>
Jean-Gabriel Young <jean-gabriel.young@uvm.edu>
"""
from collections import deque

class Layers(object):
    """Class representing partition of the nodes in layers."""

    # =========================================================================
    # Constructors
    # =========================================================================
    def __init__(self, cutpoints):
        """Initialize layers with cutpoints."""
        self._n = cutpoints[-1]
        self._max_idx = len(cutpoints)
        self._free_idx = set()
        self.layer_of = dict()
        self.layers = dict()
        u = -1
        for layer_idx, v in enumerate(cutpoints):
            self.layers[layer_idx] = deque(range(u + 1, v + 1))
            for j in range(u + 1, v + 1):
                self.layer_of[j] = layer_idx
            u = v

    @classmethod
    def from_cutpoints(cls, cutpoints):
        """Initialize layers with cutpoints.
        
        A  list of cutpoints contains the identifier of nodes at the top of
        every layer, in ascending order of identifier (i.e., descending order.
        of ranks).

        Parameters
        ----------
        cutpoints : list of int
        """
        return cls(cutpoints)

    @classmethod
    def from_dict(cls, layers_dict):
        """Initialize layers with a dictionary giving the ID of each layer.
        
        Assumes that the dictionary is well-formatted, in the sense that only
        contiguous nodes are placed in the same layer.

        Parameters
        ----------
        layers_dict: dict
            Keys are node identifiers (contiguous integers) and values are
            layer indexes.
        """
        tmp = sorted([(node, layer) for node, layer in layers_dict.items()],
                     key= lambda x: x[0])
        cutpoints = []
        current_layer = tmp[0][1]
        for node_idx, layer_idx in tmp:
            if layer_idx != current_layer:
                current_layer = layer_idx
                cutpoints.append(node_idx - 1)
        cutpoints.append(node_idx)
        return cls(cutpoints)

    @classmethod
    def from_sets(cls, layers_sets):
        """Initialize layers with a lsit of sets.

        Assumes that the list is well-formatted, in the sense that only layers
        appear in descending order with contiguous nodes placed in the same
        layer.

        Parameters
        ----------
        layers: list of sets
            List of layers encoded as sets.
        """
        cutpoints = []
        for l in layers_sets:
            cutpoints.append(max(l))
        return cls(cutpoints)

    # =========================================================================
    # Reserved methods
    # =========================================================================
    def __repr__(self):
        """Return repr(self)."""
        rep = '[' + "; ".join([str(list(v)) for l, v in self.layers.items()]) + ']'
        return rep

    def __eq__(self, other):
        """Return self==value."""
        if not isinstance(other, Layers):
            return NotImplemented
        for v in range(self._n):
            if self.layers[self.layer_of[v]] != other.layers[other.layer_of[v]]:
                return False
        return True

    def __iter__(self):
        return iter(self.as_sets())

    def __len__(self):
        return len(self.as_sets())

    # =========================================================================
    # Public methods
    # =========================================================================
    def as_sets(self):
        """Return layers as sets, in increasing order of identifiers."""
        # sort layers by node indexes
        tmp = sorted([(k, v) for (k, v) in self.layers.items()],
                     key=lambda x: x[1][0])
        # return as sets
        return [frozenset(v) for (k, v) in tmp]

    def move_up(self, v):
        """Move node v to the layer above."""
        # test that the move is possible: 
        # 1. `v` must be at the top of its layer
        # 2. v!=0
        l_v   = self.layer_of[v]
        if self.layers[l_v][0] == v and v != 0:
            l_vm1 = self.layer_of[v - 1]
            # pop out of layer and append to above layer
            self.layers[l_vm1].append(self.layers[l_v].popleft())
            # update layers
            self.layer_of[v] = l_vm1
            # check for empty layers:
            if len(self.layers[l_v]) == 0:
                del self.layers[l_v]
                self._free_idx.add(l_v)

    def move_dn(self, v):
        """Move node v to the layer below."""
        # test that the move is possible: 
        # 1. `v` must be at the bottom of its layer
        # 2. v!= n - 1
        l_v   = self.layer_of[v]
        if self.layers[l_v][-1] == v and v != self._n:
            l_vp1 = self.layer_of[v + 1]
            # pop out of layer and append to above layer
            self.layers[l_vp1].appendleft(self.layers[l_v].pop())
            # update layers
            self.layer_of[v] = l_vp1
            # check for empty layers:
            if len(self.layers[l_v]) == 0:
                del self.layers[l_v]
                self._free_idx.add(l_v)

    def split_above(self, v):
        """Split the a layer right above node v."""
        l_v   = self.layer_of[v]
        if self.layers[l_v][0] == v:
            # nothing to do
            pass
        else:
            target = self._get_free_idx()
            self.layers[target] = deque()
            if (v - self.layers[l_v][0]) < (self.layers[l_v][-1] - v):
                # fastest split is from the top
                u = None
                while u != v - 1:
                    u = self.layers[l_v].popleft()
                    self.layers[target].append(u)
                    self.layer_of[u] = target
            else:
                # fastest split is from the bottom
                u = None
                while u != v:
                    u = self.layers[l_v].pop()
                    self.layers[target].appendleft(u)
                    self.layer_of[u] = target

    def split_below(self, v):
        """Split the a layer right below node v."""
        l_v   = self.layer_of[v]
        if self.layers[l_v][-1] == v:
            # nothing to do
            pass
        else:
            target = self._get_free_idx()
            self.layers[target] = deque()
            if (v - self.layers[l_v][0]) < (self.layers[l_v][-1] - v):
                # fastest split is from the top
                u = None
                while u != v:
                    u = self.layers[l_v].popleft()
                    self.layers[target].append(u)
                    self.layer_of[u] = target
            else:
                # fastest split is from the bottom
                u = None
                while u != v + 1:
                    u = self.layers[l_v].pop()
                    self.layers[target].appendleft(u)
                    self.layer_of[u] = target

    def merge_range(self, u, v, cut=False):
        """Merge all nodes in the specified range, inclusively.

        Parameters
        ----------
        u, v: int
            Two end points of the new layer (inclusive).
        cut: bool
            If true, cut the specified range at its end to form a separated
            layers. If false, merge layers in which the endpoints are already
            included.
        """
        # index layer above and below
        if v < u:
            top = v
            bot = u
        else:
            top = u
            bot = v
        if top > 0:
            above_bot = top - 1
            above_top = min(self.layers[self.layer_of[above_bot]])
        if bot < self._n - 1:
            below_top = bot + 1
            below_bot = max(self.layers[self.layer_of[below_top]])
        # check if cuts are needed:
        if cut:
            self.split_above(top)
            self.split_below(bot)
        # merge
        start_point = min(self.layers[self.layer_of[top]]) + 1
        end_point = max(self.layers[self.layer_of[bot]])
        for v in range(start_point, end_point + 1):
            self.move_up(v)

    # =========================================================================
    # Private methods
    # =========================================================================
    def _get_free_idx(self):
        """Index manager for adding new layers."""
        if len(self._free_idx) > 0:
            return self._free_idx.pop()
        else:
            self._max_idx += 1
            return self._max_idx



# -*- coding: utf-8 -*-
"""
Partitioning in one dimension.

Authors:
Alice Patania <alice.patania@uvm.edu>
Jean-Gabriel Young <jean-gabriel.young@uvm.edu>
"""
import networkx as nx
import numpy as np
import itertools as it

class PairSumObjectives(object):
    """The abstract class of quality functions calculated as sum over pairs of nodes."""
    def __init__(self):
        self.is_pair_sum = True

    def eval_layer(self, g, upper, lower):
        """Evaluate the quality of a layer.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        upper, lower: int
            Identifiers of the nodes with the highest and lowest rank in the
            community.
        """
        f = 0
        for u, v in it.combinations_with_replacement(range(upper, lower + 1), 2):
            f += self.eval_pair(g, u, v)
        return f

    def eval(self, g, layers):
        """Evaluate the quality of a partition in layers.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        layers: 
            Either a collection of sets, in which case the sets are interpreted as
            the nodes of a layer.
        """
        Q = 0
        for L in layers:
            Q += self.eval_layer(g, min(L), max(L))
        return Q


class ArbitraryPairSumObjectives(PairSumObjectives):
    """Arbitrary PairSumObjectives quality function.

    Quality of putting pairs of nodes together specified by a matrix

    Parameters
    ----------
    f : np.array
        Quality increment array. Assumed symmetric.
    """
    def __init__(self, f=None):
        super().__init__()
        self.f = f

    def eval_pair(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph or MulDiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        # overwrite the function as soon as epsilon is set
        return self.f[u, v]


class Partitioning(PairSumObjectives):
    """Partitioning objective.
    
    Increases with the number of internal edges in layers.

    Parameters
    ----------
    penalty : float
        Strength of the penalty term.
    """
    def __init__(self, penalty=1.0):
        super().__init__()
        self.penalty = penalty

    def eval_pair(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        return g.number_of_edges(u, v) - self.penalty * ((u != v) + 1)


class Modularity(PairSumObjectives):
    """Modularity objective.
    
    Sum of the difference of the realized and expected number of edges in layers.

    Parameters
    ----------
    null_model : string
        Null model to use in the calculation.
        Options are:
            * 'CM': Configuration model
            * 'ER': Erdős–Rényi
    g : networkx DiGraph, Graph, MulDiGraph or MultiGraph (optional)
        If passed at construction, will pre-compute quantities to speed up
        calculation.
    
    Notes
    -----
    The numerical value is not normalized by the number of edges.

    """
    def __init__(self, null_model='CM', g=None):
        super().__init__()
        if null_model == 'CM':
            if g is not None:
                self.k = {u: g.degree(u) for u in g.nodes()}
                self.m = g.number_of_edges()
                self.eval_pair = self._eval_pair_CM_precompute
            else:
                self.eval_pair = self._eval_pair_CM
        if null_model == 'ER':
            self.eval_pair = self._eval_pair_ER

    def _eval_pair_CM(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        k_u = g.degree(u)
        k_v = g.degree(v)
        m = g.number_of_edges()
        return (g.number_of_edges(u, v) - k_u * k_v / (2 * m)) / ((u == v) + 1)

    def _eval_pair_CM_precompute(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        return (g.number_of_edges(u, v) - self.k[u] * self.k[v] / (2 * self.m)) / ((u == v) + 1)

    def _eval_pair_ER(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        return (g.number_of_edges(u, v) - nx.density(g)) / ((u == v) + 1)


class Egalitarian(PairSumObjectives):
    """Egalitarian objective.

    Increases with the number of reciprocated edges within layers.

    Parameters
    ----------
    g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
        Input graph used for precomputing correction terms.
    """
    def __init__(self, g=None):
        super().__init__()
        if g is None:
            self.epsilon = None
        else:
            self._set_epsilon(g)

    def _set_epsilon(self, g):
        mr = 0
        for u, v in it.combinations(g.nodes(), 2):
            if g.has_edge(u, v):
                if g.has_edge(v, u):
                    mr += 1
        n = g.number_of_nodes()
        self.epsilon = mr / (n * (n + 1) / 2)

    def eval_pair(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph or MulDiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        # overwrite the function as soon as epsilon is set
        if self.epsilon is None:
            self._set_epsilon(g)
        a_uv = g.number_of_edges(u, v)
        a_vu = g.number_of_edges(v, u)
        return a_uv * a_vu - self.epsilon


class Dominance(PairSumObjectives):
    """Dominance objective.

    Decreases with the number of unreciprocated edges within layers.

    Parameters
    ----------
    g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
        Input graph used for precomputing correction terms.
    """
    def __init__(self, g=None):
        super().__init__()
        if g is None:
            self.epsilon = None
        else:
            self._set_epsilon(g)

    def _set_epsilon(self, g):
        mu = 0
        for u, v in it.combinations(g.nodes(), 2):
            if g.has_edge(u, v):
                if not g.has_edge(v, u):
                    mu += 1
            elif g.has_edge(v, u):
                mu += 1 
        n = g.number_of_nodes()
        self.epsilon = mu / (n * (n + 1) / 2)

    def eval_pair(self, g, u, v):
        """Calculate the contribution of a pair of nodes.

        Parameters
        ----------
        g : networkx DiGraph or MulDiGraph
            Input graph.
        u, v: int
            Node identifiers.
        """
        # overwrite the function as soon as epsilon is set
        if self.epsilon is None:
            self._set_epsilon(g)
        a_uv = g.number_of_edges(u, v)
        a_vu = g.number_of_edges(v, u)
        return 2 * a_uv * a_vu - (a_uv + a_vu) + self.epsilon



# -*- coding: utf-8 -*-
"""
Utility functinos for the partitioner package.

Authors:
* Jean-Gabriel Young <jean-gabriel.young@uvm.edu>
* Alice Patania <alice.patania@uvm.edu>
"""
import networkx as nx

def preprocess(g):
    """Pre-process graph and scores.

    The nodes of the returned graph will be labeled with consecutive integers,
    in decreasing score order, and stored as a MultiGraph with collapsed nodes
    to account for equal scores.

    Parameters
    ----------
    g : networkx DiGraph, Graph, MulDiGraph or MultiGraph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    """
    sorted_scores = sorted(nx.get_node_attributes(g, 'score').items(),
                           key=lambda x: x[1],
                           reverse=True)

    # separate nodes by equivalence classes with respect to scores
    scores_to_supernode = dict()
    node_mapping = dict()
    curr_supernode = 0
    for (node_id, score) in sorted_scores:
        if score not in scores_to_supernode:
            scores_to_supernode[score] = curr_supernode
            node_mapping[curr_supernode] = [node_id]
            curr_supernode +=1
        else:
            node_mapping[scores_to_supernode[score]] += [node_id]

    # create inverse mappings 
    supernode_to_scores = dict(map(reversed, scores_to_supernode.items()))
    inv_node_mapping = dict()
    for v in node_mapping:
        for u in node_mapping[v]:
            inv_node_mapping[u] = v

    # create processed graph
    if g.is_directed():
        g_prime = nx.MultiDiGraph()
    else:
        g_prime = nx.MultiGraph()
    # add nodes
    g_prime.add_nodes_from(range(len(node_mapping)))
    nx.set_node_attributes(g_prime, node_mapping, 'node_mapping')
    nx.set_node_attributes(g_prime, supernode_to_scores, 'score')
    # add edges
    for e in g.edges():
        g_prime.add_edge(inv_node_mapping[e[0]], inv_node_mapping[e[1]])

    return g_prime



# -*- coding: utf-8 -*-
"""
Partitioning in one dimension.

Authors:
Alice Patania <alice.patania@uvm.edu>
Jean-Gabriel Young <jean-gabriel.young@uvm.edu>
"""
import networkx as nx
import numpy as np
import itertools as it
import heapq
from copy import deepcopy
from collections import defaultdict
from dyvider.layers import Layers
# from layers import Layers


# =====================================
# Private functions for heap management 
# =====================================
_REMOVED = (-np.inf, -np.inf, -np.inf, -np.inf)

def _add_move(move_heap, move_finder, move, quality=0):
    if move in move_finder:
        remove_move(move)
    entry = [quality, move]
    move_finder[move] = entry
    heapq.heappush(move_heap, entry)

def _remove_move(move_heap, move_finder, move):
    entry = move_finder.pop(move)
    quality = entry[0]
    entry[-1] = _REMOVED
    return quality

def _pop_move(move_heap, move_finder):
    while move_heap:
        quality, move = heapq.heappop(move_heap)
        if move is not _REMOVED:
            del move_finder[move]
            return quality, move

# ====================================
# Dynamical programs
# ====================================
def run(g, objective):
    """High-level wrapper for the partitioning algorithms.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Will be used to asses the quality of a partition in layers.

    See Also
    --------
    naive_dp, pair_sum_dp
    """
    try:
        if objective.is_pair_sum:
            return pair_sum_dp(g, objective)
        else:
            return naive_dp(g, objective)
    except AttributeError as e:
        return naive_dp(g, objective)

def naive_dp(g, objective):
    """Partition using an additive objective with arbitrary layer quality functions.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Will be used to asses the quality of a partition in layers. The method
        `eval_layer(g, upper, lower) -> float` should be defined and compute
        the  quality of a layer comprising nodes `upper` through `lower`
        inclusively.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    pair_sum_dp
    """
    n = g.number_of_nodes()
    Q = np.zeros(n + 1) # extra entry so that the base case is Q[-1] = 0
    optimal_cutpoint = np.zeros(n, dtype=int) # maximizing indices

    # unfolded recursion
    for j in range(0, n):
        Q[j]  = -np.inf
        for k in reversed(range(0, j + 1)):
            quality_of_split = Q[k - 1] + objective.eval_layer(g, k, j)
            if quality_of_split >  Q[j]:
                Q[j] = quality_of_split
                optimal_cutpoint[j] = k

    # retrieve cuts
    cuts = [optimal_cutpoint[-1]]
    while cuts[-1] != 0:
        cuts.append(optimal_cutpoint[cuts[-1] - 1])
    cuts = [v - 1 for v in (cuts[::-1] + [n])[1:]]

    return Layers(cuts), Q[-2]

def pair_sum_dp(g, objective):
    """Partition using an objective function with "pair-sum" quality functions.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Will be used to asses the quality of a partition in layers. The method
        `eval_pair(g, u, v) -> float` should be defined and compute the quality
        increment associated with putting nodes u and v in the same layer.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    naive_dp
    """
    n = g.number_of_nodes()
    Q = np.zeros(n + 1) # extra entry so that the base case is Q[-1] = 0
    optimal_cutpoint = np.zeros(n, dtype=int) # maximizing indices

    # create storage for layer qualities
    # implemented as a dict with default value of 0 to handle
    # boundary conditions, e.g., calls like layer_quality[(1,0)]
    # which are nonsensical and should return 0
    layer_quality = defaultdict(int)

    # unfolded recursion
    for j in range(0, n):
        Q[j]  = -np.inf
        for k in reversed(range(0, j + 1)):
            layer_quality[(k, j)] =\
                layer_quality[(k, j - 1)] +\
                layer_quality[(k + 1, j)] -\
                layer_quality[(k + 1, j - 1)] +\
                objective.eval_pair(g, k, j)
            quality_of_split = Q[k - 1] + layer_quality[(k, j)]
            if quality_of_split >  Q[j]:
                Q[j] = quality_of_split
                optimal_cutpoint[j] = k

    # retrieve top node of each layer
    bounds = [n - 1, optimal_cutpoint[-1] - 1]
    while bounds[-1] != -1:
        bounds.append(optimal_cutpoint[bounds[-1]] - 1)
    bounds = list(reversed(bounds[:-1]))  # clip end condition and reverse

    return Layers(bounds), Q[-2]

# ====================================
# Heuristics / slow programs
# ====================================

def brute_force(g, objective, num_layers=None):
    """Brute force search for the best partition.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Will be used to asses the quality of a partition in layers. The method
        `eval_layer(g, upper, lower) -> float` should be defined and compute
        the  quality of a layer comprising nodes `upper` through `lower`
        inclusively.
    num_layers : int or None
        Desired number of layers. If None, this quantity is determined
        automatically by searching over all possible number of layers.

    Warnings
    --------
    * Assumes that the graphs has been pre-processed, see utilities.py.
    * This algorithm is not designed to be efficient; it should only used as
      a benchmark.
    """
    n = g.number_of_nodes()
    # create list of all the numbers of layers that will be tried
    if num_layers is None:
        num_layers = list(range(2, n + 1))
        optimal_Q = objective.eval_layer(g, 0, n - 1)
        optimal_layers = [n - 1]
    else:
        num_layers = [num_layers]
        optimal_Q = -np.inf
        optimal_layers = None

    # search
    for k in num_layers:
        # loop over the number of layers
        # in each iteration of the loop, get boundaries of layers as
        # combinations of the node indexes
        for boundaries in it.combinations(range(n - 1), k - 1):
            Q = 0
            upper = -1
            for lower in list(boundaries) + [n - 1]:
                Q += objective.eval_layer(g, upper + 1, lower)
                upper = lower
            if Q > optimal_Q:
                optimal_Q = Q
                optimal_layers = list(boundaries) + [n - 1]
    return Layers(optimal_layers), optimal_Q


def greedy(g, objective):
    """Repeatedly apply the best local improvement starting from isolated layers.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Should have the methods:
        * eval_layer(g, upper, lower) -> float
            This functions should compute the quality of a layer comprising
            nodes `upper` through `lower` inclusively.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    """
    # initialize each node in its own communities
    n = g.number_of_nodes()
    solution = Layers.from_dict({v: v for v in range(n)})
    iteration_order = list(range(n))

    made_move = True
    while made_move:
        np.random.shuffle(iteration_order)
        made_move = False
        # iterate on nodes
        for v in iteration_order:
            # check where the node sits (top, bottom, middle of layer?)
            L = solution.layers[solution.layer_of[v]]
            top = min(L)
            bot = max(L)
            delta_Q_up = 0
            delta_Q_down = 0
            if v == top and v != 0:
                # compute the quality of moving the top node to the layer above
                above_bot = top - 1
                above_top = min(solution.layers[solution.layer_of[above_bot]])
                delta_Q_up =\
                    objective.eval_layer(g, above_top, v) -\
                    objective.eval_layer(g, above_top, above_bot) +\
                    objective.eval_layer(g, v + 1, bot) -\
                    objective.eval_layer(g, top, bot)
            if v == bot and v != n - 1:
                # compute the quality of moving the botton node to the layer below
                below_top = bot + 1
                below_bot = max(solution.layers[solution.layer_of[below_top]])
                delta_Q_down =\
                    objective.eval_layer(g, top, v - 1) -\
                    objective.eval_layer(g, top, bot) +\
                    objective.eval_layer(g, v, below_bot) -\
                    objective.eval_layer(g, below_top, below_bot)

            #  check which move is best
            merge_target = np.argmax([0, delta_Q_up, delta_Q_down])
            if merge_target == 0:
                # do nothing
                continue
            if merge_target == 1:
                # better to move the node to the layer above
                made_move = True
                solution.move_up(v)
            if merge_target == 2:
                # better to move the node to the layer below
                made_move = True
                solution.move_dn(v)

    return solution, objective.eval(g, solution)


def heap_merge(g, objective, return_trace=False):
    """Greedy algorithm inspired by Clauset, Newman and Moore (2004).

    Start each node in its own layer, and iteratively merge neighboring layers
    that would most increase (or least decrease) the objective function.
    Return the best partition encoutered during the process.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Should have the methods:
        * eval_layer(g, upper, lower) -> float
            This functions should compute the quality of a layer comprising
            nodes `upper` through `lower` inclusively.
    return_history : bool
        If true, also return the evolution of the objective and of the solution.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    pairsum_heap_merge
    """
    # initialize each node in its own layer
    n = g.number_of_nodes()
    tmp_layers = Layers.from_dict({v: v for v in range(n)})
    # initialize tracking of the best solution
    solution = deepcopy(tmp_layers)
    Q_history = [objective.eval(g, solution)]
    Q_max = Q_history[-1]
    
    move_heap = []
    move_finder = {}

    # initialize heap
    for v in range(1, n):
        above_bot = v - 1
        above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
        delta_Q =\
            objective.eval_layer(g, v - 1, v) -\
            objective.eval_layer(g, v - 1, v - 1) -\
            objective.eval_layer(g, v, v)
        _add_move(move_heap, move_finder, (v - 1, v - 1, v, v), -delta_Q)

    # run until we have a single layer
    while len(tmp_layers) > 1:
        # apply move
        (merge_delta_Q, (target_top, target_bot, top, bot)) = _pop_move(move_heap, move_finder)
        tmp_layers.merge_range(target_top, bot)

        # update solutions
        Q_history.append(Q_history[-1] - merge_delta_Q)
        if Q_history[-1] > Q_max:
            Q_max = Q_history[-1]
            solution = deepcopy(tmp_layers)

        # update merge with the layer below
        if bot < n - 1:
            below_top = bot + 1
            below_bot = max(tmp_layers.layers[tmp_layers.layer_of[below_top]])
            delta_Q =\
                objective.eval_layer(g, target_top, below_bot) -\
                objective.eval_layer(g, below_top, below_bot) -\
                objective.eval_layer(g, target_top, bot)
            _remove_move(move_heap, move_finder, (top, bot, below_top, below_bot))
            _add_move(move_heap, move_finder, (target_top, bot, below_top, below_bot), -delta_Q)
        # update merge with the layer above
        if target_top != 0:
            above_bot = target_top - 1
            above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
            delta_Q =\
                objective.eval_layer(g, above_top, bot) -\
                objective.eval_layer(g, above_top, above_bot) -\
                objective.eval_layer(g, target_top, bot)
            _remove_move(move_heap, move_finder, (above_top, above_bot, target_top, target_bot))
            _add_move(move_heap, move_finder, (above_top, above_bot, target_top, bot), -delta_Q)

    #  return
    if return_trace:
        return solution, Q_max, Q_history
    else:
        return solution, Q_max


def pair_sum_heap_merge(g, objective, return_trace=False):
    """Greedy algorithm inspired by Clauset, Newman and Moore (2004).

    Start each node in its own layer, and iteratively merge neighboring layers
    that would most increase (or least decrease) the objective function.
    Return the best partition encoutered during the process. This version of
    the function implements a speed-up specific to pair-sum objectives.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Should have the methods:
        * eval_layer(g, upper, lower) -> float
            This functions should compute the quality of a layer comprising
            nodes `upper` through `lower` inclusively.
    return_history : bool
        If true, also return the evolution of the objective and of the solution.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    heap_merge
    """
    # initialize each node in its own layer
    n = g.number_of_nodes()
    tmp_layers = Layers.from_dict({v: v for v in range(n)})
    # initialize tracking of the best solution
    solution = deepcopy(tmp_layers)
    Q_history = [objective.eval(g, solution)]
    Q_max = Q_history[-1]
    
    move_heap = []
    move_finder = {}

    # initialize heap and dictionary
    for v in range(1, n):
        above_bot = v - 1
        above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
        delta_Q =\
            objective.eval_layer(g, v - 1, v) -\
            objective.eval_layer(g, v - 1, v - 1) -\
            objective.eval_layer(g, v, v)
        _add_move(move_heap, move_finder, (v - 1, v - 1, v, v), -delta_Q)

    # run until we have a single layer
    while len(tmp_layers) > 1:
        # apply move
        (merge_delta_Q, (target_top, target_bot, top, bot)) = _pop_move(move_heap, move_finder)
        tmp_layers.merge_range(target_top, bot)

        # update solutions
        Q_history.append(Q_history[-1] - merge_delta_Q)
        if Q_history[-1] > Q_max:
            Q_max = Q_history[-1]
            solution = deepcopy(tmp_layers)

        # update merge with the layer below
        if bot < n - 1:
            below_top = bot + 1
            below_bot = max(tmp_layers.layers[tmp_layers.layer_of[below_top]])
            delta_Q = -_remove_move(move_heap, move_finder, (top, bot, below_top, below_bot))
            for u, v in it.product(range(below_top, below_bot + 1), range(target_top, target_bot + 1)):
                delta_Q += objective.eval_pair(g, u, v)
            _add_move(move_heap, move_finder, (target_top, bot, below_top, below_bot), -delta_Q)
        # update merge with the layer above
        if target_top != 0:
            above_bot = target_top - 1
            above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
            delta_Q = -_remove_move(move_heap, move_finder, (above_top, above_bot, target_top, target_bot))
            for u, v in it.product(range(top, bot + 1), range(above_top, above_bot + 1)):
                delta_Q += objective.eval_pair(g, u, v)
            _add_move(move_heap, move_finder, (above_top, above_bot, target_top, bot), -delta_Q)

    #  return
    if return_trace:
        return solution, Q_max, Q_history
    else:
        return solution, Q_max


def critical_gap(g, objective, return_trace=False):
    """Clusters nodes greedily based on proximity in the embedding space.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Should have the methods:
        * eval_layer(g, upper, lower) -> float
            This functions should compute the quality of a layer comprising
            nodes `upper` through `lower` inclusively.
    return_history : bool
        If true, also return the evolution of the objective and of the solution.

    Notes
    -----
    Proceeds by merging the closest pairs of nodes into layers instead of by
    literally changing the gap. This is equivalent but much faster to the
    gap-variation method.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    pair_sum_critical_gap
    """
    # initialize each node in its own layer
    n = g.number_of_nodes()
    tmp_layers = Layers.from_dict({v: v for v in range(n)})
    # initialize tracking of the best solution
    solution = deepcopy(tmp_layers)
    Q_history = [objective.eval(g, solution)]
    Q_max = Q_history[-1]

    move_heap = []
    move_finder = {}

    # initialize heap
    for v in range(1, n):
        above_bot = v - 1
        above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
        delta_Q =\
            objective.eval_layer(g, v - 1, v) -\
            objective.eval_layer(g, v - 1, v - 1) -\
            objective.eval_layer(g, v, v)
        distance = abs(g.nodes[v]['score'] - g.nodes[v-1]['score'])
        _add_move(move_heap, move_finder, (v - 1, v - 1, v, v), (distance, -delta_Q))

    # run until we have a single layer
    while len(tmp_layers) > 1:
        # apply move
        ((distance, merge_delta_Q), (target_top, target_bot, top, bot)) = _pop_move(move_heap, move_finder)
        tmp_layers.merge_range(target_top, bot)

        # update solutions
        Q_history.append(Q_history[-1] - merge_delta_Q)
        if Q_history[-1] > Q_max:
            Q_max = Q_history[-1]
            solution = deepcopy(tmp_layers)

        # update merge with the layer below
        if bot < n - 1:
            below_top = bot + 1
            below_bot = max(tmp_layers.layers[tmp_layers.layer_of[below_top]])
            delta_Q =\
                objective.eval_layer(g, target_top, below_bot) -\
                objective.eval_layer(g, below_top, below_bot) -\
                objective.eval_layer(g, target_top, bot)
            distance = abs(g.nodes[bot]['score'] - g.nodes[below_top]['score'])
            _remove_move(move_heap, move_finder, (top, bot, below_top, below_bot))
            _add_move(move_heap, move_finder, (target_top, bot, below_top, below_bot), (distance, -delta_Q))
        # update merge with the layer above
        if target_top != 0:
            above_bot = target_top - 1
            above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
            delta_Q =\
                objective.eval_layer(g, above_top, bot) -\
                objective.eval_layer(g, above_top, above_bot) -\
                objective.eval_layer(g, target_top, bot)
            distance = abs(g.nodes[above_bot]['score'] - g.nodes[target_top]['score'])
            _remove_move(move_heap, move_finder, (above_top, above_bot, target_top, target_bot))
            _add_move(move_heap, move_finder, (above_top, above_bot, target_top, bot), (distance, -delta_Q))

    #  return
    if return_trace:
        return solution, Q_max, Q_history
    else:
        return solution, Q_max


def pair_sum_critical_gap(g, objective, return_trace=False):
    """Clusters nodes greedily based on proximity in the embedding space.

    This version of the function implements a speed up specific to pair-sum
    objectives.

    Parameters
    ----------
    g : networkx DiGraph or Graph
        Annotated graph structure. Nodes should have a "score" attribute
        corresponding to their embedding position.
    objective : objective function object
        Should have the methods:
        * eval_layer(g, upper, lower) -> float
            This functions should compute the quality of a layer comprising
            nodes `upper` through `lower` inclusively.
    return_history : bool
        If true, also return the evolution of the objective and of the solution.

    Notes
    -----
    Proceeds by merging the closest pairs of nodes into layers instead of by
    literally changing the gap. This is equivalent but much faster to the
    gap-variation method.

    Warnings
    --------
    Assumes that the graphs has been pre-processed, see utilities.py.

    See Also
    --------
    critical_gap
    """
    # initialize each node in its own layer
    n = g.number_of_nodes()
    tmp_layers = Layers.from_dict({v: v for v in range(n)})
    # initialize tracking of the best solution
    solution = deepcopy(tmp_layers)
    Q_history = [objective.eval(g, solution)]
    Q_max = Q_history[-1]

    move_heap = []
    move_finder = {}

    # initialize heap
    for v in range(1, n):
        above_bot = v - 1
        above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
        delta_Q =\
            objective.eval_layer(g, v - 1, v) -\
            objective.eval_layer(g, v - 1, v - 1) -\
            objective.eval_layer(g, v, v)
        distance = abs(g.nodes[v]['score'] - g.nodes[v-1]['score'])
        _add_move(move_heap, move_finder, (v - 1, v - 1, v, v), (distance, -delta_Q))

    # run until we have a single layer
    while len(tmp_layers) > 1:
        # apply move
        ((distance, merge_delta_Q), (target_top, target_bot, top, bot)) = _pop_move(move_heap, move_finder)
        tmp_layers.merge_range(target_top, bot)

        # update solutions
        Q_history.append(Q_history[-1] - merge_delta_Q)
        if Q_history[-1] > Q_max:
            Q_max = Q_history[-1]
            solution = deepcopy(tmp_layers)

        # update merge with the layer below
        if bot < n - 1:
            below_top = bot + 1
            below_bot = max(tmp_layers.layers[tmp_layers.layer_of[below_top]])

            (_, priority) = _remove_move(move_heap, move_finder, (top, bot, below_top, below_bot))
            delta_Q = -priority
            for u, v in it.product(range(below_top, below_bot + 1), range(target_top, target_bot + 1)):
                delta_Q += objective.eval_pair(g, u, v)
            distance = abs(g.nodes[bot]['score'] - g.nodes[below_top]['score'])
            _add_move(move_heap, move_finder, (target_top, bot, below_top, below_bot), (distance, -delta_Q))
        # update merge with the layer above
        if target_top != 0:
            above_bot = target_top - 1
            above_top = min(tmp_layers.layers[tmp_layers.layer_of[above_bot]])
            (_, priority) = _remove_move(move_heap, move_finder, (above_top, above_bot, target_top, target_bot))
            delta_Q = -priority
            for u, v in it.product(range(top, bot + 1), range(above_top, above_bot + 1)):
                delta_Q += objective.eval_pair(g, u, v)
            distance = abs(g.nodes[above_bot]['score'] - g.nodes[target_top]['score'])
            _add_move(move_heap, move_finder, (above_top, above_bot, target_top, bot), (distance, -delta_Q))

    #  return
    if return_trace:
        return solution, Q_max, Q_history
    else:
        return solution, Q_max



"""
dyvider

Authors:
Alice Patania <alice.patania@uvm.edu>
Jean-Gabriel Young <jean-gabriel.young@uvm.edu>
"""
import pkg_resources

from . import (
    utilities,
    objectives,
    algorithms,
    layers
)


__version__ = pkg_resources.require("dyvider")[0].version


import time
import datetime
from setuptools import setup

# Version information
name = 'dyvider'
major = "0"
minor = "1"

description="Fast and exact partitioning algorithms for graphs embedded in one dimension."

license = 'MIT'
authors = "Jean-Gabriel Young, Alice Patania"
author_email = "jean-gabriel.young@uvm.edu"
url = 'https://github.com/jg-you/dyvider'
download_url = 'https://pypi.python.org/pypi/dyvider/'
platforms = ['Linux', 'Unix']
keywords = ['graphs', 'community detection', 'networks', 'inference']
classifiers = [
    'Development Status :: 2 - Pre-Alpha  ',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License  ',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Scientific/Engineering :: Information Analysis',
    'Topic :: Scientific/Engineering :: Mathematics']

date_info = datetime.datetime.utcfromtimestamp(int(time.time()))
date = time.asctime(date_info.timetuple())
___version___ = major + "." + minor


setup(name=name,
      version=___version___,
      description=description,
      long_description=open('long_description.rst').read(),
      url=url,
      author=authors,
      author_email=author_email,
      classifiers=classifiers,
      keywords=keywords,
      license=license,
      test_suite='nose.collector',
      tests_require=['nose'],
      install_requires=[
          'networkx',
          'numpy'
      ])